#include "arch.h"

#if 0 && defined(__L3_PPCASM__)

#warning Using assembly version of window_dual

/*
 PowerPC assembly version of functions from cwin.c

 This version of the file was written with Mac OS X Server in mind, but it
 should not be difficult to get this to work with Linux PPC.
 */



// This should be defined by the calling Makefile if the target chip
// supports the 'fsel' instruction.  I'm mostly interested in the PPC750
// so I haven't built branchless (or less branchy) versions of the integer
// bounds checking (its not to get it down to one branch).
#ifdef PPC_HAS_FSEL
#warning Enabling usage of the 'fsel' instruction
#else
#warning Disabling usage of the 'fsel' instruction
#endif


// Since we have plenty of registers, we'll statically assign
// them (this also makes the code easier to read below).  Available
// registers on MacOS X Server are r0, r3-r12, fr0-fr13 and anything else
// we want to save (but these should be enough for this simple routine).
// r0 means '0' in a lot of cases so it isn't very convenient to use.

// inputs
#define VBUF   r3
#define VB_PTR r4
#define PCM    r5

// temporaries
#define SI      r6
#define BX      r7
#define COEF    r8
#define TMP     r9
#define PICBASE r10
#define WINCOEF r11

#define ZERO      f0
#define SUMA      f1
#define SUMB      f2
#define FTMP0A    f3
#define FTMP0B    f4
#define FTMP1A    f5
#define FTMP1B    f6
#ifdef PPC_HAS_FSEL
#define SHRT_MIN  f7
#define SHRT_MAX  f8
#endif // PPC_HAS_FSEL

// Stack space for converting from float->int
#define CONVERT_DOUBLE   -8(r1)   // a 64-bit aligned slot
#define CONVERT_SINGLE   -4(r1)   // the part that will contain the resulting int


//
// Mangling for names when using C++
//
#ifdef __cplusplus
#define NAME  _window_dual_ppc__FPCfiPs
#else
#define NAME  _window_dual_ppc
#endif


.data
.const
	.align 2
zero:
	.single 0.0
#ifdef PPC_HAS_FSEL
shrt_min:
	.single -32768.0
shrt_max:
	.single 32767.0
#endif // PPC_HAS_FSEL


.text
	.align 2

////
////
////  Non-scheduled version.  This is the reference copy of the assembly version.  Changes should be
////  made to this version and then any instruction-level scheduled versions should be placed below.
////  This makes it easier to see and maintain the transformation from the C code to the assembly.
////  (Also, you can then use automated tools to take the unscheduled version (or blocks thereof)
////  and schedule them.
////
////

#define NON_SCHEDULED

#ifdef NON_SCHEDULED

.globl NAME
NAME:
	// We need to play PIC games and have local space.  We won't build a frame
        // since we aren't going to exceed the red zone.
	mflr	TMP		// save the link register
	bl	L_picbase	// put the addr of the next insn in the link reg
L_picbase:
	mflr	PICBASE		// get our PIC base address
	mtlr	TMP		// restore the link register


	// si   = vb_ptr + 16;
	// bx   = (si + 32) & 511;
	// si *= sizeof(float);
	// bx *= sizeof(float);
	addi	SI, VB_PTR, 16
	addi	BX, VB_PTR, 32+16
        slwi    SI, SI, 2
        rlwinm  BX, BX, 2, 21, 29

	// coef = wincoef;
	addis	TMP,PICBASE,ha16(L_wincoef$non_lazy_ptr - L_picbase)
	addi	TMP,TMP,lo16(L_wincoef$non_lazy_ptr - L_picbase)
	lwz	WINCOEF,0(TMP)
	mr	COEF,WINCOEF

	// load constants
	addis	TMP,PICBASE,ha16(zero - L_picbase)
	lfs	ZERO,lo16(zero - L_picbase)(TMP)

#ifdef PPC_HAS_FSEL
	addis	TMP,PICBASE,ha16(shrt_min - L_picbase)
	lfs	SHRT_MIN,lo16(shrt_min - L_picbase)(TMP)
	
	addis	TMP,PICBASE,ha16(shrt_max - L_picbase)
	lfs	SHRT_MAX,lo16(shrt_max - L_picbase)(TMP)
#endif

	// First 16
	li	TMP,16
	mtctr	TMP
L_first:
	fmr	SUMA, ZERO
	fmr	SUMB, ZERO
	
	// coef[0] and coef[1]
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,0(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,4(COEF)
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	// coef[2] and coef[3]
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,8(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,12(COEF)
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	// coef[4] and coef[5]
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,16(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,20(COEF)
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	// coef[6] and coef[7]
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,24(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,28(COEF)
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	// coef[8] and coef[9]
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,32(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,36(COEF)
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	// coef[10] and coef[11]
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,40(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,44(COEF)
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	// coef[12] and coef[13]
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,48(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,52(COEF)
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	// coef[14] and coef[15]
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,56(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,60(COEF)
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	addi	COEF,COEF,16*4
	addi	SI,SI,4
	addi	BX,BX,-4

	fadds	SUMA,SUMA,SUMB

#ifdef PPC_HAS_FSEL
	// Clip to SHRT_MIN..SHRT_MAX (using fsel to avoid branches)
	fsubs	FTMP0A,SUMA,SHRT_MIN	// tmp0 = sum + 32768
	fsubs	FTMP1A,SUMA,SHRT_MAX	// tmp1 = sum - 32767
	fsel	SUMA,FTMP0A,SUMA,SHRT_MIN  // if (sum < -32768) sum = -32768;
	fsel	SUMA,FTMP1A,SHRT_MAX,SUMA  // if (sum >= 32767) sum = 32767;
#endif

	fctiwz	FTMP0A,SUMA
	stfd	FTMP0A, CONVERT_DOUBLE
	lwz	TMP, CONVERT_SINGLE

#ifndef PPC_HAS_FSEL
	cmpwi	TMP, 32767
	ble	L_LessThan0
	li	TMP,32767
	b	L_GreaterThan0
L_LessThan0:
	cmpwi	TMP, -32768
	bgt	L_GreaterThan0
	li	TMP,-32768
L_GreaterThan0:
#endif

	sth	TMP,0(PCM)
	addi	PCM,PCM,4
	

	// Decrement the loop counter and repeat if not zero
	bdnz    L_first


	// special case
	fmr	SUMA,ZERO
	fmr	SUMB,ZERO

	lfsx	FTMP0A,BX,VBUF
	lfs	FTMP1A,0(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31

	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,4(COEF)
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31

	lfsx	FTMP0A,BX,VBUF
	lfs	FTMP1A,8(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31

	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,12(COEF)
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31

	lfsx	FTMP0A,BX,VBUF
	lfs	FTMP1A,16(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31

	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,20(COEF)
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31

	lfsx	FTMP0A,BX,VBUF
	lfs	FTMP1A,24(COEF)
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31

	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,28(COEF)
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31

	addi	COEF,COEF,8*4
	fadds	SUMA,SUMA,SUMB

#ifdef PPC_HAS_FSEL
	// Clip to SHRT_MIN..SHRT_MAX (using fsel to avoid branches)
	fsubs	FTMP0A,SUMA,SHRT_MIN	// tmp0 = sum + 32768
	fsubs	FTMP1A,SUMA,SHRT_MAX	// tmp1 = sum - 32767
	fsel	SUMA,FTMP0A,SUMA,SHRT_MIN  // if (sum < -32768) sum = -32768;
	fsel	SUMA,FTMP1A,SHRT_MAX,SUMA  // if (sum >= 32767) sum = 32767;
#endif

	fctiwz	FTMP0A,SUMA
	stfd	FTMP0A, CONVERT_DOUBLE
	lwz	TMP, CONVERT_SINGLE

#ifndef PPC_HAS_FSEL
	cmpwi	TMP, 32767
	ble	L_LessThan1
	li	TMP,32767
	b	L_GreaterThan1
L_LessThan1:
	cmpwi	TMP, -32768
	bgt	L_GreaterThan1
	li	TMP,-32768
L_GreaterThan1:
#endif

	sth	TMP,0(PCM)
	addi	PCM,PCM,4


	// Last 15
	addi	COEF,WINCOEF,255*4
	li	TMP,15
	mtctr	TMP

L_last:
	addi	SI,SI,-4
	addi	BX,BX,4

	fmr	SUMA,ZERO
	fmr	SUMB,ZERO

	// 0 and -1
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,0(COEF)
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,-4(COEF)
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	// -2 and -3
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,-8(COEF)
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,-12(COEF)
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	// -4 and -5
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,-16(COEF)
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,-20(COEF)
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	// -6 and -7
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,-24(COEF)
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,-28(COEF)
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	// -8 and -9
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,-32(COEF)
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,-36(COEF)
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	// -10 and -11
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,-40(COEF)
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,-44(COEF)
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	// -12 and -13
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,-48(COEF)
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,-52(COEF)
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	// -14 and -15
	lfsx	FTMP0A,SI,VBUF
	lfs	FTMP1A,-56(COEF)
	addi	SI,SI,64*4
	rlwinm	SI,SI,0,21,31
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF
	lfs	FTMP1B,-60(COEF)
	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB


	addi	COEF,COEF,-16*4
	fadds	SUMA,SUMA,SUMB

#ifdef PPC_HAS_FSEL
	// Clip to SHRT_MIN..SHRT_MAX (using fsel to avoid branches)
	fsubs	FTMP0A,SUMA,SHRT_MIN	// tmp0 = sum + 32768
	fsubs	FTMP1A,SUMA,SHRT_MAX	// tmp1 = sum - 32767
	fsel	SUMA,FTMP0A,SUMA,SHRT_MIN  // if (sum < -32768) sum = -32768;
	fsel	SUMA,FTMP1A,SHRT_MAX,SUMA  // if (sum >= 32767) sum = 32767;
#endif

	fctiwz	FTMP0A,SUMA
	stfd	FTMP0A, CONVERT_DOUBLE
	lwz	TMP, CONVERT_SINGLE

#ifndef PPC_HAS_FSEL
	cmpwi	TMP, 32767
	ble	L_LessThan2
	li	TMP,32767
	b	L_GreaterThan2
L_LessThan2:
	cmpwi	TMP, -32768
	bgt	L_GreaterThan2
	li	TMP,-32768
L_GreaterThan2:
#endif
	sth	TMP,0(PCM)
	addi	PCM,PCM,4

	// Decrement the loop counter and repeat if not zero
	bdnz    L_last

	// Done!
	blr	

#else  // NON_SCHEDULED

////
////
////  Scheduled version with FSEL
////
////
#ifndef PPC_HAS_FSEL
#error No scheduled version without 'fsel' support
#endif

.globl NAME
NAME:
	// We need to play PIC games and have local space.  We won't build a frame
        // since we aren't going to exceed the red zone.
	mflr	TMP		// save the link register
	bl	L_picbase	// put the addr of the next insn in the link reg
L_picbase:
	mflr	PICBASE		// get our PIC base address
	mtlr	TMP		// restore the link register


	// si   = vb_ptr + 16;
	// bx   = (si + 32) & 511;
	// si *= sizeof(float);
	// bx *= sizeof(float);
	addi	SI, VB_PTR, 16
	addi	BX, VB_PTR, 32+16
        slwi    SI, SI, 2
        rlwinm  BX, BX, 2, 21, 29

	// coef = wincoef;
	addis	TMP,PICBASE,ha16(L_wincoef$non_lazy_ptr - L_picbase)
	addi	TMP,TMP,lo16(L_wincoef$non_lazy_ptr - L_picbase)
	lwz	WINCOEF,0(TMP)
	mr	COEF,WINCOEF

	// load constants
	addis	TMP,PICBASE,ha16(zero - L_picbase)
	lfs	ZERO,lo16(zero - L_picbase)(TMP)

#ifdef PPC_HAS_FSEL
	addis	TMP,PICBASE,ha16(shrt_min - L_picbase)
	lfs	SHRT_MIN,lo16(shrt_min - L_picbase)(TMP)
	
	addis	TMP,PICBASE,ha16(shrt_max - L_picbase)
	lfs	SHRT_MAX,lo16(shrt_max - L_picbase)(TMP)
#endif

	// First 16
	li	TMP,16
	mtctr	TMP
L_first:
	lfsx	FTMP0B,BX,VBUF
	fmr	SUMA, ZERO

	lfsx	FTMP0A,SI,VBUF
	addi	SI,SI,64*4

	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	lfs	FTMP1A,0(COEF)
	// Empty

	lfs	FTMP1B,4(COEF)
	fmr	SUMB, ZERO

	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA

	lfs	FTMP1A,8(COEF)
	// Empty

	lfsx	FTMP0A,SI,VBUF
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfs	FTMP1B,12(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,16(COEF)

	lfsx	FTMP0A,SI,VBUF
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfs	FTMP1B,20(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,24(COEF)

	lfsx	FTMP0A,SI,VBUF
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfs	FTMP1B,28(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,32(COEF)

	lfsx	FTMP0A,SI,VBUF
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfs	FTMP1B,36(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,40(COEF)

	lfsx	FTMP0A,SI,VBUF
	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	lfs	FTMP1B,44(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,48(COEF)

	lfsx	FTMP0A,SI,VBUF
	addi	SI,SI,64*4

	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	lfs	FTMP1B,52(COEF)

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	lfsx	FTMP0A,SI,VBUF
	addi	SI,SI,64*4

	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047
	rlwinm	SI,SI,0,21,31  // and with 511*4+4-1 = 2047

	lfs	FTMP1A,56(COEF)
	addi	SI,SI,4

	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	lfs	FTMP1B,60(COEF)

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	rlwinm	BX,BX,0,21,31  // and with 511*4+4-1 = 2047

	addi	COEF,COEF,16*4
	addi	BX,BX,-4

	fnmsubs SUMB,FTMP0B,FTMP1B,SUMB
	// Empty

	fadds	SUMA,SUMA,SUMB
	fsubs	FTMP0A,SUMA,SHRT_MIN	// tmp0 = sum + 32768
	fsubs	FTMP1A,SUMA,SHRT_MAX	// tmp1 = sum - 32767
	fsel	SUMA,FTMP0A,SUMA,SHRT_MIN  // if (sum < -32768) sum = -32768;
	fsel	SUMA,FTMP1A,SHRT_MAX,SUMA  // if (sum >= 32767) sum = 32767;

	fctiwz	FTMP0A,SUMA
	stfd	FTMP0A, CONVERT_DOUBLE
	lwz	TMP, CONVERT_SINGLE
	addi	PCM,PCM,4
	sth	TMP,-4(PCM)

	// Decrement the loop counter and repeat if not zero
	bdnz    L_first


	// special case
	lfs	FTMP1A,0(COEF)
	fmr	SUMA,ZERO

	lfsx	FTMP0A,BX,VBUF
	// Empty

	fmr	SUMB,ZERO
	addi	BX,BX,64*4

	lfs	FTMP1B,4(COEF)
	rlwinm	BX,BX,0,21,31

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF

	lfs	FTMP1A,8(COEF)
	addi	BX,BX,64*4

	// STALL
	rlwinm	BX,BX,0,21,31

	fmadds	SUMB,FTMP0B,FTMP1B,SUMB
	lfsx	FTMP0A,BX,VBUF

	lfs	FTMP1B,12(COEF)
	addi	BX,BX,64*4

	// STALL
	rlwinm	BX,BX,0,21,31

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF

	addi	BX,BX,64*4
	// Empty

	lfs	FTMP1A,16(COEF)
	rlwinm	BX,BX,0,21,31

	fmadds	SUMB,FTMP0B,FTMP1B,SUMB
	lfsx	FTMP0A,BX,VBUF

	addi	BX,BX,64*4
	// Empty

	lfs	FTMP1B,20(COEF)
	rlwinm	BX,BX,0,21,31

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF

	addi	BX,BX,64*4
	// Empty

	lfs	FTMP1A,24(COEF)
	rlwinm	BX,BX,0,21,31

	fmadds	SUMB,FTMP0B,FTMP1B,SUMB
	lfsx	FTMP0A,BX,VBUF

	lfs	FTMP1B,28(COEF)
	addi	COEF,COEF,8*4

	addi	BX,BX,64*4
	rlwinm	BX,BX,0,21,31

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfsx	FTMP0B,BX,VBUF

	addi	BX,BX,64*4
	// Empty

	fmadds	SUMB,FTMP0B,FTMP1B,SUMB
	rlwinm	BX,BX,0,21,31
	fadds	SUMA,SUMA,SUMB

	// Clip to SHRT_MIN..SHRT_MAX (using fsel to avoid branches)
	fsubs	FTMP0A,SUMA,SHRT_MIN	// tmp0 = sum + 32768
	fsubs	FTMP1A,SUMA,SHRT_MAX	// tmp1 = sum - 32767
	fsel	SUMA,FTMP0A,SUMA,SHRT_MIN  // if (sum < -32768) sum = -32768;
	fsel	SUMA,FTMP1A,SHRT_MAX,SUMA  // if (sum >= 32767) sum = 32767;

	fctiwz	FTMP0A,SUMA
	stfd	FTMP0A, CONVERT_DOUBLE
	lwz	TMP, CONVERT_SINGLE


	sth	TMP,0(PCM)
	addi	PCM,PCM,4


	// Last 15
	addi	COEF,WINCOEF,255*4
	li	TMP,15
	mtctr	TMP

L_last:
	addi	BX,BX,4
	addi	SI,SI,-4

	fmr	SUMA,ZERO
	lfsx	FTMP0B,BX,VBUF

	lfsx	FTMP0A,SI,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31

	lfs	FTMP1A,0(COEF)
	// Empty

	lfs	FTMP1B,-4(COEF)
	fmr	SUMB,ZERO

	rlwinm	SI,SI,0,21,31
	fmadds	SUMA,FTMP0A,FTMP1A,SUMA

	lfs	FTMP1A,-8(COEF)
	// Empty

	lfsx	FTMP0A,SI,VBUF
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31

	rlwinm	SI,SI,0,21,31
	lfs	FTMP1B,-12(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,-16(COEF)

	lfsx	FTMP0A,SI,VBUF
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31

	rlwinm	SI,SI,0,21,31
	lfs	FTMP1B,-20(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,-24(COEF)

	lfsx	FTMP0A,SI,VBUF
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31

	rlwinm	SI,SI,0,21,31
	lfs	FTMP1B,-28(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,-32(COEF)

	lfsx	FTMP0A,SI,VBUF
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31

	rlwinm	SI,SI,0,21,31
	lfs	FTMP1B,-36(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,-40(COEF)

	lfsx	FTMP0A,SI,VBUF
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31

	rlwinm	SI,SI,0,21,31
	lfs	FTMP1B,-44(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,-48(COEF)

	lfsx	FTMP0A,SI,VBUF
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31

	rlwinm	SI,SI,0,21,31
	lfs	FTMP1B,-52(COEF)

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	lfs	FTMP1A,-56(COEF)

	lfsx	FTMP0A,SI,VBUF
	fmadds	SUMB,FTMP0B,FTMP1B,SUMB

	lfsx	FTMP0B,BX,VBUF
	addi	BX,BX,64*4

	addi	SI,SI,64*4
	rlwinm	BX,BX,0,21,31

	fmadds	SUMA,FTMP0A,FTMP1A,SUMA
	rlwinm	SI,SI,0,21,31

	lfs	FTMP1B,-60(COEF)
	// Empty

	fmadds	SUMB,FTMP0B,FTMP1B,SUMB
	addi	COEF,COEF,-16*4
	fadds	SUMA,SUMA,SUMB

	// Clip to SHRT_MIN..SHRT_MAX (using fsel to avoid branches)
	fsubs	FTMP0A,SUMA,SHRT_MIN	// tmp0 = sum + 32768
	fsubs	FTMP1A,SUMA,SHRT_MAX	// tmp1 = sum - 32767
	fsel	SUMA,FTMP0A,SUMA,SHRT_MIN  // if (sum < -32768) sum = -32768;
	fsel	SUMA,FTMP1A,SHRT_MAX,SUMA  // if (sum >= 32767) sum = 32767;

	fctiwz	FTMP0A,SUMA
	stfd	FTMP0A, CONVERT_DOUBLE
	lwz	TMP, CONVERT_SINGLE

	sth	TMP,0(PCM)
	addi	PCM,PCM,4

	// Decrement the loop counter and repeat if not zero
	bdnz    L_last

	// Done!
	blr	


#endif // NON_SCHEDULED


.data
.non_lazy_symbol_pointer
L_wincoef$non_lazy_ptr:
	.indirect_symbol _wincoef
	.long 0

#endif
